from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.rpc import Base
import logging
import math
import re
import time
from webfriend import exceptions

RX_ID_EXPANSION = re.compile('#(?P<id>[^\s]+)')


class NoSuchElement(Exception):
    pass


class InvalidElement(Exception):
    pass


class DOMElement(object):
    def __init__(self, rpc, definition):
        self.dom = rpc
        self.populate(definition)

    def populate(self, definition):
        if not isinstance(definition, dict):
            raise AttributeError("DOM Element definition must be a dict")

        if 'nodeId' not in definition or definition['nodeId'] is None:
            raise AttributeError("DOM Element definition must contain 'nodeId'")

        if 'nodeName' not in definition:
            self._partial = True
        else:
            self._partial = False

        self.definition  = definition
        self.id          = definition['nodeId']
        self.parent_id   = definition.get('parentId')
        self.type        = definition.get('nodeType')
        self.local_name  = definition.get('localName')
        self.attributes  = {}
        self._box_model  = None
        self._name       = definition.get('nodeName')
        self._value      = definition.get('nodeValue')
        self.child_nodes = []
        self._text       = ''
        self.child_ids   = set()
        self._bounding_rect = None

        # populate children
        if len(definition.get('children', [])):
            for child in definition['children']:
                child_type = child.get('nodeType')

                if child_type == 1:
                    self.child_ids.add(child['nodeId'])

                elif child_type == 3:
                    self._text = child.get('nodeValue', '')

                else:
                    self.child_nodes.append(DOMElement(self.dom, child))

        # populate attributes
        if len(definition.get('attributes', [])):
            attrgen = (a for a in definition['attributes'])

            for name, value in [(n, attrgen.next()) for n in attrgen]:
                self.attributes[name] = value

    def evaluate(self, script, return_by_value=False, own_properties=False, accessors_only=True):
        remote_object = self.dom.call('resolveNode', nodeId=self.id).result.get('object', {})
        object_id = remote_object['objectId']

        # retrieve details via script injection
        try:
            # call the function to retrieve runtime position info
            result = self.dom.tab.runtime.call_function_on(
                object_id,
                "function(){{ {} }}".format(script),
                return_by_value=return_by_value,
            ).result.get('result')

            # if we asked to return the value, then do so now
            if return_by_value:
                return result

            result_id = result.get('objectId')

            if result_id:
                # retrieve the object that resulted from that call
                properties = self.dom.tab.runtime.get_properties(
                    result_id,
                    own_properties=own_properties,
                    accessors_only=accessors_only
                )

                # release that object
                self.dom.tab.runtime.release_object(result_id)
            else:
                properties = []

            # return the data
            return dict([
                (p['name'], p.get('value', {}).get('value')) for p in properties if not p['name'].startswith('_')
            ])

        finally:
            if not return_by_value:
                self.dom.tab.runtime.release_object(object_id)

    def click(self, ensure_target=True, scroll_to=True):
        if self.attributes.get('target') == '_blank':
            del self['target']

        # if specified, scroll down to the element if it's below the fold
        if scroll_to:
            if self.top > self.dom.root.height:
                self.scroll_to()

        return self.evaluate("return this.click()", return_by_value=True)

    def scroll_to(self, x=None, y=None):
        if x is None:
            x = 0

        if y is None:
            y = self.top

        return self.evaluate(
            "window.scrollTo({}, {})".format(x, y),
            return_by_value=True
        )

    @property
    def is_partial(self):
        return self._partial

    @property
    def name(self):
        return self._name

    # @name.setter
    # def name(self, v):
    #     response = self.dom.call('setNodeName', nodeId=self.id, name='{}'.format(v))
    #     return response.get('params', {})['nodeId']

    @property
    def content_document(self):
        if isinstance(self.definition.get('contentDocument'), dict):
            if 'nodeId' in self.definition['contentDocument']:
                return DOMElement(self.dom, self.definition['contentDocument'])

        return None

    @property
    def text(self):
        return self._text

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self.dom.call('setNodeValue', nodeId=self.id, value='{}'.format(v))

    @property
    def outer_html(self):
        return self.dom.call('getOuterHTML', nodeId=self.id).get('outerHTML')

    @outer_html.setter
    def outer_html(self, value):
        self.call('setOuterHTML', nodeId=self.id, outerHTML=value)

    def refresh_attributes(self):
        pairs = self.dom.call('getAttributes', nodeId=self.id).get('attributes', [])
        attrgen = (a for a in pairs)
        attrs = {}

        for name, value in [(n, attrgen.next()) for n in attrgen]:
            attrs[name] = value

        self.attributes = attrs

        return self.attributes

    @property
    def children(self):
        out = []

        for i in self.child_ids:
            el = self.dom.element(i)

            if el:
                out.append(el)

        return out

    def recheck_box(self):
        self._box_model = None
        return self.box

    def recheck_bounds(self):
        self._bounding_rect = None
        return self.bounds

    @property
    def bounds(self):
        if self._bounding_rect is None:
            self._bounding_rect = self.evaluate(
                "return this.getBoundingClientRect();",
                return_by_value=True
            )

        return self._bounding_rect

    @property
    def box(self):
        if self._box_model is None:
            self._box_model = self.dom.call('getBoxModel', nodeId=self.id).get('model')

        return self._box_model

    @property
    def parent(self):
        return self.dom.element(self.parent_id)

    @property
    def width(self):
        return int(math.ceil(self.bounds.get('width', self.box['width'])))

    @property
    def height(self):
        return int(math.ceil(self.bounds.get('height', self.box['height'])))

    @property
    def top(self):
        return self.bounds.get('top')

    @property
    def right(self):
        return self.bounds.get('right')

    @property
    def bottom(self):
        return self.bounds.get('bottom')

    @property
    def left(self):
        return self.bounds.get('left')

    def query(self, selector):
        return self.dom.query(selector, node_id=self.id)

    def query_all(self, selector):
        return self.dom.query_all(selector, node_id=self.id)

    def xpath(self, expression):
        """
        Query the DOM for elements matching the given XPath expression, starting
        from the current element.

        #### Arguments

        - **expression** (`str`):

            An XPath expression to evaluate within the current DOM.

        #### Returns
        A list of matching elements.
        """
        return self.dom.xpath(expression, node_id=self.id)

    def remove(self):
        self.dom.remove(self.id)

    def focus(self):
        self.dom.focus(self.id)

    def __getitem__(self, key):
        return self.attributes[key]

    def __contains__(self, key):
        return (key in self.attributes)

    def __setitem__(self, key, value):
        self.dom.call('setAttributeValue', nodeId=self.id, name=key, value=str(value))

    def __delitem__(self, key):
        self.dom.call('removeAttribute', nodeId=self.id, name=key)

    def __repr__(self):
        if self.name:
            if self.type == 1:
                element_name = self.name.lower()
                attrs = []

                for k, v in self.attributes.items():
                    attrs.append('{}="{}"'.format(k, v))

                if len(attrs):
                    attrs = ' ' + ' '.join(attrs)
                else:
                    attrs = ''

                return '<{}{}>{}</{}>'.format(
                    element_name,
                    attrs,
                    self.text,
                    element_name
                )
            else:
                return '{} id={}'.format(self.name, self.id)
        else:
            parts = ['node={}'.format(self.id)]

            if self.name:
                parts.append('name={}'.format(self.name))

            return 'DOMElement<{}>'.format(','.join(parts))

    def __str__(self):
        return self.__repr__()


class DOM(Base):
    domain = 'DOM'
    _you_never_forget_your_first_root_element = None
    _root_element = None
    _elements = {}

    def initialize(self):
        self.on('setChildNodes', self.on_child_nodes)
        self.on('childNodeInserted', self.on_child_inserted)
        self.on('childNodeRemoved', self.on_child_removed)
        self.tab.page.on('frameClearedScheduledNavigation', self.reset)

    def reset(self, *args, **kwargs):
        """
        Clears out any existing local definitions in the DOM tree. This is used when navigating
        between pages so that the new remote DOM can be stored.
        """
        self._root_element = None

        if kwargs.get('clear_initial_root', True):
            self._you_never_forget_your_first_root_element = None

    def clear_requests(self, *args, **kwargs):
        self.tab.reset_network_request_cache()

    def has_element(self, id):
        """
        Return whether we know about the given element ID or not.

        #### Arguments

        - **id** (`str`):

            The DOM element id.
        """
        if id in self._elements:
            return True
        return False

    def element(self, id, fallback=None):
        """
        Return the given ID as a `webfriend.rpc.dom.DOMElement`, or `None` if we have not
        received the element from Chrome.

        #### Arguments

        - **id** (`str`):

            The DOM element id.

        - **fallback** (any):

            The value to return if the element is not known.

        #### Returns
        - `webfriend.rpc.dom.DOMElement` if the element is known, **fallback** if not.
        """
        return self._elements.get(id, fallback)

    def element_at(self, x, y, shadow=False):
        """
        Return the topmost `webfriend.rpc.dom.DOMElement` at the given page coordinates, or `None`
        if there is no element there.

        #### Arguments

        - **x** (`int`):

            The X-coordinate to test at.

        - **y** (`int`):

            The Y-coordinate to test at.

        - **shadow** (`bool`):

            If specified, include "shadow DOMs" in the query; see
            https://developers.google.com/web/fundamentals/getting-started/primers/shadowdom

        #### Returns
        - `webfriend.rpc.dom.DOMElement` if an element is at the coordinates, `None` if not.
        """
        node_id = self.call(
            'getNodeForLocation',
            x=x,
            y=y,
            includeUserAgentShadowDOM=shadow
        ).get('nodeId')

        if node_id:
            return self.element(node_id)

        return None

    def print_node(self, node_id=None, level=0, indent=4):
        """
        Print details about the given element to the debug log.
        """
        if node_id is None:
            node_id = self.root.id

        node = self.element(node_id)

        if node is not None:
            logging.debug('TREE: {}{}'.format((' ' * indent * level), node))

            for child in node.children:
                self.print_node(child.id, level=level + 1)

    @property
    def root(self):
        """
        Returns the root DOM element for the current document.

        #### Returns
        The `webfriend.rpc.dom.DOMElement` representing the root element.
        """
        if self._root_element is None:
            self.reset()

            self._root_element = DOMElement(
                self,
                self.call('getDocument', depth=1, pierce=True).get('root')
            )

            # keep track of the VERY first root element encountered, so that if we switch roots
            # for traversing frames, we can always come back to this one
            if self._you_never_forget_your_first_root_element is None:
                self._you_never_forget_your_first_root_element = self._root_element

            self._elements[self._root_element.id] = self._root_element

        return self._root_element

    @property
    def scroll_width(self):
        return self.root.evaluate(
            "return document.documentElement.scrollWidth",
            return_by_value=True
        ).get('value', None)

    @property
    def scroll_height(self):
        return self.root.evaluate(
            "return document.documentElement.scrollHeight",
            return_by_value=True
        ).get('value', None)

    @property
    def frames(self):
        return self.query_all('frame')

    def root_to(self, selector):
        elements = self.select_nodes(selector, wait_for_match=True)

        if elements:
            element = self.ensure_unique_element(selector, elements['nodes'])

            subdoc = element.content_document

            if subdoc is not None:
                # reset, we've got a document, but leave the top-level document in place
                # so we can go back to it
                self.reset(clear_initial_root=False)

                self.call('requestChildNodes', nodeId=subdoc.id)
                self._root_element = subdoc
                logging.debug('DOM root is now {}'.format(self.root))
                return self.root
            else:
                raise InvalidElement("Element does not contain a nested document")

        raise NoSuchElement("No elements matched the query selector '{}'".format(selector))

    def root_to_top(self):
        if self._you_never_forget_your_first_root_element:
            self._root_element = self._you_never_forget_your_first_root_element

        logging.debug('DOM root is now {}'.format(self.root))
        return self.root

    def remove_node(self, node_id):
        """
        Remove a specific node (and all children) from the DOM.

        - **id** (`str`):

            The DOM element id.
        """
        self.call('removeNode', nodeId=node_id)

    def focus(self, node_id):
        """
        Set the focus on a specific element.

        - **id** (`str`):

            The DOM element id.
        """
        self.call('focus', nodeId=node_id)

    def query(self, selector, node_id=None, reply_timeout=None):
        """
        Query for a single element.  Matched elements will be populated in the local DOM element
        cache for rapid retrieval in subsequent calls.

        - **selector** (`str`):

            The page element to query for, given as a CSS-style selector, an ID (e.g. "#myid"), or
            an XPath query (e.g.: "xpath://body/p").

        - **node_id** (`str`):

            If specified, the query will be performed relative to the element with the given
            Node ID.  Otherwise, the root element will be used.

        - **reply_timeout** (`int`):

            The timeout (in milliseconds) before the query raises a
            `webfriend.exceptions.TimeoutError`.

        #### Returns
        The matching `webfriend.rpc.dom.DOMElement`.

        #### Raises
        - `webfriend.exceptions.EmptyResult` if zero elements were matched, or
        - `webfriend.exceptions.TooManyResults` if more than one elements were matched.
        - `webfriend.exceptions.TimeoutError` if the query times out.
        """
        if node_id is None:
            node_id = self.root.id

        # perform the call
        matched_node_id = self.call(
            'querySelector',
            reply_timeout=reply_timeout,
            selector=self.prepare_selector(selector),
            nodeId=node_id
        ).get('nodeId')

        if not matched_node_id:
            raise NoSuchElement(
                "No elements matched the query selector '{}' under document {}".format(
                    selector,
                    node_id
                )
            )

        # we try to retrieve the element from the local element cache because they typically
        # come in as a series of events before the querySelector reply comes back, meaning that
        # by the time we get to this point we _should_ have cached definitions
        #
        # if we don't have a cached copy, then return a skeleton instance so that we can still
        # perform operations on the returned element
        #
        return self.element(matched_node_id, DOMElement(self, {
            'nodeId': matched_node_id,
        }))

    def query_all(self, selector, node_id=None, reply_timeout=None):
        """
        Query for all matching elements. Matched elements will be populated in the local DOM element
        cache for rapid retrieval in subsequent calls.

        - **selector** (`str`):

            The page element to query for, given as a CSS-style selector, an ID (e.g. "#myid"), or
            an XPath query (e.g.: "xpath://body/p").

        - **node_id** (`str`):

            If specified, the query will be performed relative to the element with the given
            Node ID.  Otherwise, the root element will be used.

        - **reply_timeout** (`int`):

            The timeout (in milliseconds) before the query raises a
            `webfriend.exceptions.TimeoutError`.

        #### Returns
        A `list` of matching `webfriend.rpc.dom.DOMElement` (zero or more).

        #### Raises
        - `webfriend.exceptions.TimeoutError` if the query times out.
        """
        if node_id is None:
            node_id = self.root.id

        # perform the call
        node_ids = self.call(
            'querySelectorAll',
            reply_timeout=reply_timeout,
            selector=self.prepare_selector(selector),
            nodeId=node_id
        ).get('nodeIds', [])

        # we try to retrieve the element from the local element cache because they typically
        # come in as a series of events before the querySelector reply comes back, meaning that
        # by the time we get to this point we _should_ have cached definitions
        #
        # if we don't have a cached copy, then return a skeleton instance so that we can still
        # perform operations on the returned element
        #
        return [
            self.element(i, DOMElement(self, {
                'nodeId': i,
            })) for i in node_ids
        ]

    def xpath(self, expression, timeout=10000):
        """
        Query the DOM for elements matching the given XPath expression.

        #### Arguments

        - **expression** (`str`):

            An XPath expression to evaluate within the current DOM.

        #### Returns
        A list of matching elements.
        """

        # call getDocument first so subsequent calls to performSearch work properly
        # (thanks @jamcplusplus for catching this)
        self.root

        # perform the call
        handle = self.perform_search(expression)
        count = handle.get('resultCount', 0)

        if count:
            return self.get_search_results(handle.get('searchId'), to_index=count)
        else:
            self.discard_search_results(handle.get('searchId'))
            return []

    def select_nodes(self, selector, wait_for_match=True, timeout=10000, interval=250):
        """
        Polls the DOM for an element that matches the given selector.  Either the element will be
        found and returned within the given timeout, or a TimeoutError will be raised.

        #### Arguments

        - **selector** (`str`):

            The CSS-style selector that specifies the DOM element to look for.

        - **timeout** (`int`):

            The number of milliseconds to wait for the element to be returned.

        - **interval** (`int`):

            The polling interval, in milliseconds, used for rechecking for the element.

        #### Returns
        `webfriend.rpc.dom.DOMElement`

        #### Raises
        `webfriend.exceptions.TimeoutError`
        """
        started_at = time.time()

        while time.time() <= (started_at + (timeout / 1e3)):
            try:
                elements = self.query_all(selector, reply_timeout=interval)

                # if we're not waiting or we are and we've got elements, return
                if not wait_for_match or len(elements):
                    results = {
                        'selector': selector,
                        'count':    len(elements),
                        'nodes': [
                            self.element(e.id, e) for e in elements
                        ],
                    }

                    if not len(results['nodes']):
                        return False

                    return results
            except (
                exceptions.TimeoutError,
                exceptions.ProtocolError
            ):
                pass

            time.sleep(interval / 1e3)

        return False

    def on_child_nodes(self, event):
        for node in event.get('nodes', []):
            element = DOMElement(self, node)

            logging.debug('Adding node {} {}'.format(
                element.id,
                element
            ))

            self._elements[node['nodeId']] = element

            if self.has_element(element.parent_id):
                self.element(element.parent_id).child_ids.add(element.id)

    def on_child_inserted(self, event):
        definition = event.get('node')
        parent_id = event.get('parentNodeId')

        element = DOMElement(self, definition)
        self._elements[element.id] = element

        logging.debug('Inserting node {} {}'.format(
            element.id,
            element
        ))

        if self.has_element(parent_id):
            try:
                self.element(parent_id).child_ids.add(element.id)
            except KeyError:
                pass

    def on_child_removed(self, event):
        node_id = event.get('nodeId')
        parent_id = event.get('parentNodeId')

        if self.has_element(parent_id):
            try:
                self.element(parent_id).child_ids.remove(node_id)
            except KeyError:
                pass

        if self.has_element(node_id):
            logging.debug('Removing node {} {}'.format(
                self._elements[node_id].id,
                self._elements[node_id]
            ))

            del self._elements[node_id]

    # def on_set_attribute_value()

    @property
    def resources(self):
        return self.tab._network_requests

    def get_resource(self, url=None, request_id=None):
        """
        Return a `dict` describing a specific network resource, or false if none was found.

        #### Arguments

        - **url** (`str`, optional):

            If specified, this is the URL of the requested resource to retrieve.

        - **request_id** (`str`, optional):

            If specified, this is the specific resource to retrieve by Request ID.

        #### Returns
        A `dict` representing the resource, or _false_ if a match was not found.
        """
        if not request_id and not url:
            raise ValueError("Must specify either url or request_id")

        if request_id:
            return self.tab.get_network_request(request_id)
        elif url:
            matching_requests = []

            for _, request in self.resources.items():
                try:
                    if url == request['request']['url']:
                        return request
                    elif url in request['request']['url']:
                        matching_requests.append(request)
                except KeyError:
                    continue

            if len(matching_requests) == 1:
                return matching_requests[0]
            elif len(matching_requests) > 1:
                raise exceptions.TooManyResults(
                    'The URL pattern "{}" matched {} requests, need at most 1'.format(
                        url,
                        len(matching_requests)
                    )
                )

        return False

    @classmethod
    def prepare_selector(cls, selector):
        selector = RX_ID_EXPANSION.sub('*[id*="\g<id>"]', selector)

        return selector

    @classmethod
    def ensure_unique_element(cls, selector, elements):
        if elements is False:
            raise exceptions.EmptyResult("No elements matched the selector: {}".format(
                selector
            ))

        if isinstance(elements, dict):
            nodes = elements.get('nodes', [])
        else:
            nodes = elements

        if not len(nodes):
            raise exceptions.EmptyResult("No elements matched the selector: {}".format(
                selector
            ))

        elif len(nodes) > 1:
            raise exceptions.TooManyResults(
                "Selector {} is ambiguous, matched {} elements".format(
                    selector, len(nodes)
                )
            )

        return nodes[0]

    def perform_search(self, query, shadow=True):
        params = {
            'query': query,
        }

        if shadow:
            params['includeUserAgentShadowDOM'] = shadow

        return self.call('performSearch', **params)

    def get_search_results(self, search_id, from_index=0, to_index=0):
        node_ids = self.call(
            'getSearchResults',
            searchId=search_id,
            fromIndex=from_index,
            toIndex=to_index
        ).get('nodeIds', [])

        return [
            self.element(i, DOMElement(self, {
                'nodeId': i,
            })) for i in node_ids
        ]

    def discard_search_results(self, search_id):
        self.call('discardSearchResults', searchId=search_id)

    def push_node_by_path_to_frontend(self, path):
        node_ids = self.call('pushNodeByPathToFrontend', path=path).get('nodeIds', [])

        return [
            self.element(i, DOMElement(self, {
                'nodeId': i,
            })) for i in node_ids
        ]

    def push_node_by_backend_ids_to_frontend(self, backend_node_ids):
        return self.call('pushNodesByBackendIdsToFrontend', backendNodeIds=backend_node_ids)

    def request_node(self, remote_object):
        return self.call('requestNode', objectId=remote_object).get('nodeId')



from __future__ import absolute_import
from webfriend.rpc import Base
from uuid import uuid4
import json


class EmbeddedScriptError(Exception):
    def __init__(self, message, line=None, col=None, **kwargs):
        self.line = line
        self.col = col
        super(Exception, self).__init__(message, **kwargs)


class Runtime(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Runtime
    """
    domain = 'Runtime'
    supports_events = False

    def evaluate(
        self,
        expression,
        object_group=None,
        include_command_line_api=True,
        silent=None,
        context_id=None,
        return_by_value=True,
        generate_preview=None,
        user_gesture=None,
        await_promise=None,
        obj_own_properties=True,
        obj_accessors_only=False,
        obj_preview=False,
        wrapper_fn=True,
        data=None,
        calling_context=None
    ):
        if isinstance(data, dict):
            data_serialized = json.dumps(data)
        else:
            data_serialized = '{}'  # empty javascript object literal

        if wrapper_fn:
            fn_name = 'webfriend_{}_fn'.format(str(uuid4()).replace('-', '_'))

            # declare variable to hold the anonymous function, bound to the data that was passed in
            # and call the function
            expression = 'var {} = function(){{ {} }}.bind({}); {}();'.format(
                fn_name,
                expression,
                data_serialized,
                fn_name,
                fn_name
            )

        params = {
            'expression':    expression,
            'returnByValue': return_by_value,
        }

        if object_group is not None:
            params['objectGroup'] = object_group

        if include_command_line_api is not None:
            params['includeCommandLineAPI'] = include_command_line_api

        if silent is not None:
            params['silent'] = silent

        if context_id is not None:
            params['contextId'] = context_id

        if generate_preview is not None:
            params['generatePreview'] = generate_preview

        if user_gesture is not None:
            params['userGesture'] = user_gesture

        if await_promise is not None:
            params['awaitPromise'] = await_promise

        reply = self.call('evaluate', **params)

        if 'exceptionDetails' in reply.result:
            exception   = reply.result['exceptionDetails']
            detail      = exception.get('exception', {})
            line_number = exception['lineNumber'] + 1
            col_number  = exception['columnNumber'] + 1  # add 1 because javascript columns are zero-based, we're not

            if calling_context:
                if hasattr(calling_context, 'line'):
                    line_number += calling_context.line

            # provide LOTS of context from both where the injected script failed AND where
            # the script was injected from in our local call stack.
            raise EmbeddedScriptError('{}'.format(
                detail.get(
                    'description',
                    detail.get(
                        'value',
                        exception.get(
                            'text',
                            'Unknown Error'
                        )
                    )
                ),
                line_number
            ), line=line_number, col=col_number)

        else:
            result = reply.result.get('result', {})

            if return_by_value:
                return result.get('value')
            elif 'objectId' in result:
                return self.get_properties(
                    result['objectId'],
                    own_properties=obj_own_properties,
                    accessors_only=obj_accessors_only,
                    preview=obj_preview
                )
            else:
                return None

    def call_function_on(self, object_id, fn_name, arguments=None, return_by_value=False):
        return self.call(
            'callFunctionOn',
            objectId=object_id,
            functionDeclaration=fn_name,
            arguments=[{
                'value': v,
            } for v in (arguments or [])],
            returnByValue=return_by_value
        )

    def get_properties(
        self,
        object_id,
        own_properties=False,
        accessors_only=False,
        preview=False
    ):
        params = {
            'objectId': object_id,
        }

        if own_properties:
            params['ownProperties'] = True

        if accessors_only:
            params['accessorPropertiesOnly'] = True

        if preview:
            params['generatePreview'] = True

        return self.call('getProperties', **params).get('result')

    def release_object(self, object_id):
        self.call('releaseObject', objectId=object_id)

    def release_object_group(self, object_group):
        self.call('releaseObjectGroup', objectGroup=object_group)



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.commands.base import CommandProxy
from webfriend.scripting.scope import Scope


class StateProxy(CommandProxy):
    qualifier = 'vars'

    def default(self, parent=0):
        """
        Return a list of variable names that are defined in a scope.

        #### Arguments

        - **parent** (`int`, optional):

            If non-zero, specifies how many levels up to insepect relative to the current scope.
            This is used to query and manipulate scopes above our own.

        #### Returns
        A list of zero of more strings, one for each variable name.
        """
        return self._scope_at_level(parent).as_dict().keys()

    def interpolate(self, format, isolated=False, values=None):
        """
        Return a value interpolated with values from a scope or ones that are explicitly provided.

        #### Arguments

        - **format** (`str`):

            The format string to interpolate.

        - **isolated** (`bool`):

            If true, the calling scope will not be used to interpolate values.

        - **values** (`dict`, optional):

            If specified, these values will override the scope's values (if any) and be available
            to **format** during interpolation.

        #### Returns
        An autotyped value resulting from interpolating the given **format**.
        """
        if isolated:
            scope = Scope()
        else:
            scope = Scope(parent=self.scope)

        if isinstance(values, dict):
            for k, v in values.items():
                scope[k] = v

        return self.environment.interpolate(format, scope=scope)

    def get(self, key, fallback=None, parent=0):
        """
        Return the value of a specific variable defined in a scope.

        #### Arguments

        - **key** (`str`):

            The name of the variable to retrieve.

        - **fallback** (`bool`, optional):

            A value to return if the variable was not found.

        - **parent** (`int`, optional):

            If non-zero, specifies how many levels up to insepect relative to the current scope.
            This is used to query and manipulate scopes above our own.

        #### Returns
        The value of the named variable, or **fallback** if the value was not found.
        """
        return self._scope_at_level(parent).get(key, fallback)

    def set(self, key, value, interpolate=True, parent=0):
        if interpolate is True:
            value = self.interpolate(value)

        self._scope_at_level(parent)[key] = value
        return value

    def clear(self, key, parent=0):
        scope = self._scope_at_level(parent)

        if key in scope:
            del scope[key]
            return True

        return False

    def push(self, key, value, interpolate=True, parent=0):
        scope = self._scope_at_level(parent)

        if key not in scope:
            scope[key] = []

        if not isinstance(scope[key], list):
            scope[key] = [scope[key]]

        if interpolate is True:
            value = self.interpolate(value)

        scope[key].append(value)
        return value

    def pop(self, key, parent=0):
        scope = self._scope_at_level(parent)

        if key in scope:
            if isinstance(scope[key], list):
                try:
                    return scope[key].pop()
                except IndexError:
                    pass

        return None

    def ensure(self, key, parent=0, message=None):
        if message is None:
            message = '{key} must be specified.'

        try:
            scope = self._scope_at_level(parent)
            return scope[key]
        except KeyError:
            raise KeyError(message.format(**{
                'key':   key,
                'scope': scope,
            }))

    def _scope_at_level(self, level=0):
        if not level:
            return self.scope
        else:
            if level is True:
                level = 1

            ancestors = self.scope.ancestors()
            index = len(ancestors) + (-1 * (level + 1))

            if index < 0:
                index = 0

            return ancestors[index]



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.rpc import (
    Base,
    Browser,
    Console,
    DOM,
    Emulation,
    Input,
    Network,
    Overlay,
    Page,
    Reply,
    Runtime,
    Target,
)
import json
import time
from webfriend import exceptions
from webfriend.utils import patch_json  # noqa
import websocket
import logging
from Queue import Queue, Empty, Full
from threading import Thread

ANY_KEY = 'ANY'


class Tab(object):
    default_width  = 0
    default_height = 0

    def __init__(
        self,
        browser,
        description,
        domains=None,
        width=None,
        height=None,
        frame_id=None,
        callbacks=True,
        autoresize=True
    ):
        if not isinstance(description, dict):
            raise AttributeError("Tab descriptor must be a dict")

        if 'webSocketDebuggerUrl' not in description:
            raise AttributeError("Cannot operate on tab without a webSocketDebuggerUrl")

        if width is None:
            width = self.default_width

        if height is None:
            height = self.default_height

        self.browser           = browser
        self.frame_id          = frame_id
        self.description       = description
        self.message_id        = 0
        self.socket            = websocket.create_connection(self.wsurl)
        self.waiters           = {}
        self.triggerqueue      = Queue()
        self.last_event_m      = {}
        self.last_event_t      = {}
        self._network_requests = {}
        self.g_recv_ctl        = Queue(1)
        self.g_recv            = Thread(target=self.receive_messages, args=(self.g_recv_ctl,))
        self.replies           = {}
        self.initial_w         = width
        self.initial_h         = height
        self.msg_enable        = False
        self.netreq_tracking   = True
        self._trigger_worker   = None

        # setup and enable all the RPC domains we support
        self.page              = Page(self)
        self.dom               = DOM(self)
        self.console           = Console(self)
        self.emulation         = Emulation(self)
        self.input             = Input(self)
        self.network           = Network(self)
        self.runtime           = Runtime(self)
        self.window            = Browser(self)
        self.overlay           = Overlay(self)
        self.target            = Target(self)

        # start the receive thread
        self.g_recv.start()

        for domain in self.rpc_domains:
            domain.initialize()

        # setup internal callbacks
        if callbacks:
            self.setup_callbacks()

        # perform initial calls
        if autoresize:
            if self.initial_w or self.initial_h:
                self.emulation.set_device_metrics_override(
                    width=self.initial_w,
                    height=self.initial_h,
                )

    @property
    def url(self):
        return self.description.get('url')

    @property
    def wsurl(self):
        return self.description.get('webSocketDebuggerUrl')

    @property
    def rpc_domains(self):
        instances = []

        for k in dir(self):
            if k == 'rpc_domains':
                continue

            attr = getattr(self, k)

            if isinstance(attr, Base):
                instances.append(attr)

        return instances

    def as_dict(self):
        return {
            'id': self.frame_id,
            'url': self.url,
            'webSocketDebuggerUrl': self.wsurl,
            'target': (self.frame_id == self.browser.default_tab),
        }

    def enable_events(self):
        for domain in self.rpc_domains:
            domain.enable()

    def enable_console_messages(self):
        self.msg_enable = True

    def disable_console_messages(self):
        self.msg_enable = False

    def enable_network_request_tracking(self):
        self.netreq_tracking = True

    def disable_network_request_tracking(self):
        self.netreq_tracking = False

    def stop(self):
        if self.g_recv.is_alive():
            logging.debug('Sending stop to receive thread')
            self.g_recv_ctl.put(StopIteration)
            self.socket.close()

        while self.g_recv.is_alive():
            logging.debug('Waiting for receive thread...')
            time.sleep(1)

    def send(self, data, expect_reply=True, reply_timeout=None, context=None):
        if not isinstance(data, dict):
            raise AttributeError("Data must be a dict")

        if not reply_timeout:
            reply_timeout = 10000

        # increment and include message ID
        self.message_id += 1
        data['id'] = self.message_id

        body = json.dumps(data)

        try:
            request_handle = {
                'id':    data['id'],
                'reply': Queue(1),
            }

            self.replies[data['id']] = request_handle

            # send the request to the Remote Debugger
            logging.debug(' >> [{:04d}] {} {}'.format(
                data['id'],
                data['method'],
                ' '.join([
                    '{}={}'.format(k, v) for k, v in data.get('params', {}).items()
                ])
            ))

            # send the request
            self.socket.send(body)

            # block until the receive loop says so
            if expect_reply:
                try:
                    reply, events = request_handle['reply'].get(timeout=(reply_timeout / 1e3))
                except Empty:
                    raise exceptions.TimeoutError("Timed out waiting for reply to command '{}', id={}".format(
                        data['method'],
                        data['id']
                    ))

                # if there was an exception, raise it now
                if isinstance(reply, Exception):
                    raise reply

                # make sure the IDs match
                if reply['id'] == data['id']:
                    return Reply(reply, request=data, events=events)
                else:
                    raise exceptions.ProtocolError("Reply Message ID does not match Request Message ID")

            else:
                return None

        finally:
            del self.replies[data['id']]

    def dispatch_event(self, message):
        if message is StopIteration:
            logging.info('Sending stop to trigger thread')
            self.triggerqueue.put((None, None, StopIteration))
        else:
            domain, method = message.get('method').split('.', 1)
            payload = message.get('params', {})

            try:
                proxy = self.get_domain_instance(domain)
            except ValueError:
                logging.exception('Unhandled Event Type')
                return

            self.triggerqueue.put((proxy, method, payload))

    def trigger_worker(self):
        while True:
            proxy, method, payload = self.triggerqueue.get()

            if payload is StopIteration:
                logging.debug('Stopping trigger thread')
                return

            event = proxy.trigger(method, payload)
            event_name = str(event)

            if event:
                logging.debug(' >> [ .. ] EVENT: {}'.format(
                    event
                ))

                # record the current time as the last time we saw an event of this type
                now = time.time()
                self.last_event_m[str(event)] = now
                self.last_event_t = now

                # attempt to send this event to whoever is waiting for it
                if event_name in self.waiters:
                    try:
                        self.waiters[event_name].put(event)
                    except Full:
                        pass

                if ANY_KEY in self.waiters:
                    try:
                        self.waiters[ANY_KEY].put(event)
                    except Full:
                        pass

    def dispatch_reply(self, request_id, message, events):
        if request_id in self.replies:
            self.replies[request_id]['reply'].put((message, events))
        else:
            logging.warning('Received message without a sender (id={})'.format(request_id))

    def receive_messages(self, controlq):
        self._trigger_worker = Thread(target=self.trigger_worker)
        self._trigger_worker.start()

        while True:
            try:
                try:
                    if controlq.get_nowait() is StopIteration:
                        raise
                except Empty:
                    pass

                message = self.receive()

                if message is None:
                    continue

                # print(json.dumps(message, indent=4))

                if isinstance(message, Exception):
                    self.dispatch_reply(message.id, message, [])

                elif 'id' in message:
                    self.dispatch_reply(message['id'], message, [])

                else:
                    self.dispatch_event(message)

            except (KeyboardInterrupt, StopIteration, websocket.WebSocketException) as e:
                logging.debug('Fatal receive message: {}'.format(e))
                break

        self.dispatch_event(StopIteration)
        logging.info('Stopping receive thread')

    def receive(self, timeout=10):
        message = self.socket.recv()

        if message is not None:
            body = json.loads(message)
            exc = None

            if 'error' in body:
                if isinstance(body['error'], dict):
                    error = body['error']

                    message = error.get('message', 'Unknown Error')

                    if 'data' in error:
                        message += ' - {}'.format(error['data'])

                    exc = exceptions.ProtocolError(
                        'Protocol Error {}: {}'.format(error.get('code', -1), message)
                    )
                else:
                    exc = exceptions.ProtocolError('Malformed Error Response')

            if exc is not None:
                exc.id = body.get('id')
                return exc

            return body

        else:
            return None

    def wait_for_caller_response(self, event_name, timeout=30000):
        """
        Yields events of
        Block until a specific event is received, or until **timeout** elapses (whichever comes first).

        #### Arguments

        - **event_name** (`str`):

            The name of the event to wait for.

        - **timeout** (`int`):

            The timeout, in milliseconds, before raising a `webfriend.exceptions.TimeoutError`.

        #### Returns
        `webfriend.rpc.event.Event`

        #### Raises
        `webfriend.exceptions.TimeoutError`
        """

        # get or create the async result for this event
        if event_name not in self.waiters:
            result = Queue(1)
            self.waiters[event_name] = result
        else:
            result = self.waiters[event_name]

        try:
            event = result.get(timeout=(timeout / 1e3))
            accepted = yield event

            # generator received a response from the caller
            if accepted is not None:
                return

        except Empty:
            raise exceptions.TimeoutError("Timed out waiting for events")

        finally:
            del self.waiters[event_name]

    def wait_for(self, event_name, **kwargs):
        """
        Block until a specific event is received, or until **timeout** elapses (whichever comes first).

        #### Arguments

        - **event_name** (`str`):

            The name of the event to wait for.

        - **timeout** (`int`):

            The timeout, in milliseconds, before raising a `webfriend.exceptions.TimeoutError`.

        #### Returns
        `webfriend.rpc.event.Event`

        #### Raises
        `webfriend.exceptions.TimeoutError`
        """
        wfc = self.wait_for_caller_response(event_name, **kwargs)
        started_at = time.time()

        for event in wfc:
            try:
                wfc.send(True)
            except StopIteration:
                pass

            return {
                'sequence': [event],
                'duration': (time.time() - started_at),
            }

    def wait_for_idle(self, idle, events=[], timeout=30000, poll_interval=250):
        """
        Blocks for a specified amount of time _after_ an event has been received, or until
        **timeout** elapses (whichever comes first).

        This is useful for waiting for events to occur after performing an action, then giving some
        amount of time for those events to "settle" (e.g.: allowing the page time to react to those
        events without knowing ahead of time what, if any, listeners will be responding.)  A common
        use case for this would be to wait a few seconds _after_ a resize has occurred for anything
        that just loaded to finish doing so.


        #### Arguments

        - **idle** (`int`):

            The amount of time, in milliseconds, that the event stream should be idle before
            returning.

        - **events** (`list`, optional):

            If not empty, the **idle** time will be interpreted as the amount of time since _any
            of these specific events_ have occurred.  The default is to wait for the browser to be
            idle with respect to _any_ events.

        - **timeout** (`int`):

            The maximum amount of time to wait before raising a
            `webfriend.exceptions.TimeoutError`.

        - **poll_interval** (`int`):

            How often to check the event timings to see if the idle time has elapsed.

        #### Returns
        An `int` representing the number of milliseconds we waited for.

        #### Raises
        `webfriend.exceptions.TimeoutError`
        """
        started_at = time.time()
        idle = (idle / 1e3)

        # clear out the old timings we're interested in
        if len(events):
            for name in events:
                if name in self.last_event_m:
                    del self.last_event_m[name]

        while time.time() < (started_at + (timeout / 1e3)):
            # if we don't have an event filter, then we just want to wait idle seconds
            # after ANY event has been received
            if not len(events):
                # if the time since ANY event has met/exceeded our idle time
                if time.time() >= (self.last_event_t + idle):
                    return (time.time() - started_at) * 1e3

            else:
                for name in events:
                    if name in self.last_event_m:
                        # if the time since this event has met/exceeded our idle time
                        if time.time() >= (self.last_event_m[name] + idle):
                            # now that we've gotten the event, remove it so subsequent calls wait
                            # for the next one to occur
                            del self.last_event_m[name]
                            return (time.time() - started_at) * 1e3

            time.sleep(poll_interval / 1e3)

        raise exceptions.TimeoutError("Timed out waiting for events to stop coming in")

    def evaluate(self, *args, **kwargs):
        return self.runtime.evaluate(*args, **kwargs)

    def rpc(self, method, expect_reply=True, reply_timeout=None, context=None, **params):
        payload = {
            'method': method,
        }

        if len(params):
            payload['params'] = params

        return self.send(
            payload,
            expect_reply=expect_reply,
            reply_timeout=reply_timeout,
            context=context
        )

    def get_domain_instance(self, domain):
        for attr in dir(self):
            instance = getattr(self, attr)

            if isinstance(instance, Base):
                if instance.domain == domain:
                    return instance

        raise ValueError("No such instance for domain '{}'".format(domain))

    def on(self, event_pattern, callback):
        domain, _ = event_pattern.split('.', 1)
        instance = self.get_domain_instance(domain)
        return instance.on(event_pattern, callback)

    def remove_handler(self, callback_id):
        domain, _ = callback_id.split('.', 1)
        instance = self.get_domain_instance(domain)
        return instance.remove_handler(callback_id)

    def reset_network_request_cache(self):
        self._network_requests = {}

    def get_network_request(self, request_id):
        return self._network_requests.get(request_id)

    def setup_callbacks(self):
        def on_net_pre_request(e):
            self._network_requests[e.get('requestId')] = {
                'before': e,
            }

        def on_net_response_received(e):
            nid = e.get('requestId')

            if isinstance(self._network_requests.get(nid), dict):
                self._network_requests[nid]['success'] = True
                self._network_requests[nid]['completed'] = True
                self._network_requests[nid]['response'] = e

        def on_net_load_failed(e):
            nid = e.get('requestId')

            if isinstance(self._network_requests.get(nid), dict):
                self._network_requests[nid]['success'] = False
                self._network_requests[nid]['response'] = e

        def on_message(e):
            message = e.get('message', {})
            level = message.get('level', 'log')
            body = message.get('text', '').strip()

            if isinstance(body, str):
                body = body.encode('UTF-8')

            if len(body):
                l = '--'

                if level == 'info':
                    l = 'II'
                elif level == 'warning':
                    l = 'WW'
                elif level == 'error':
                    l = 'EE'
                elif level == 'debug':
                    l = 'DD'

                logging.info('[{}] {}'.format(l, body))

        if self.netreq_tracking:
            self.network.on('requestWillBeSent', on_net_pre_request)
            self.network.on('responseReceived',  on_net_response_received)
            self.network.on('loadingFailed',     on_net_load_failed)
            logging.debug('Network request tracking is enabled')
        else:
            logging.debug('Network request tracking is disabled')

        if self.msg_enable:
            self.console.on('messageAdded', on_message)



from __future__ import absolute_import
from __future__ import unicode_literals
from unittest import TestCase
from webfriend.scripting import parser
from webfriend.scripting.execute import execute_script


class ScriptingTest(TestCase):
    def setUp(self):
        self.maxDiff = 10000

    def _eval(self, script):
        return execute_script(None, script).as_dict()

    def test_assignment(self):
        self.assertEqual({
            'null': None,
            'a': 1,
            'b': True,
            'c': "Test",
            'c1': "Test Test",
            'c2': "Test {c}",
            'd': 3.14159,
            'e': [
                1,
                True,
                "Test",
                3.14159,
                [
                    1,
                    True,
                    "Test",
                    3.14159,
                ],
                {
                    'ok': True,
                }
            ],
            'f': {
                'ok': True,
            },
            'g': 'g',
            'h': 'h',
            'i': 'i',
            'j': 'j',
            'k': None,
            'l': 'l',
            'm': 'm',
            'o': 'o',
            'p': None,
            'q': 'q',
            's': 's',
            't': True,
            'u': None,
            'v': 1,
            'w': True,
            'x': "Test",
            'y': 3.14159,
            'z': [1, True, "Test", 3.14159],
            'z0': 1,
            'z1': True,
            'z2': 'Test',
            'z3': 3.14159,
            'aa': 1,
            'bb': True,
            'cc': "Test",
            'dd': 3.14159,
            'ee': {
                'ok': True,
                'always': {
                    'finishing': {
                        'each_others': 'sentences',
                    },
                },
            },
            'ee1': {
                'finishing': {
                    'each_others': 'sentences',
                },
            },
            'ee2': {
                'each_others': 'sentences',
            },
            'ee3': 'sentences',
            'ee4': None,
            'ee5': True,
            'ee6': 'sentences',
            'ekey': 'always',
            'ee7': 'sentences',
            'ee8': {
                'ok': True,
                'always': {
                    'finishing': {
                        'each_others': 'sandwiches',
                        'other': {
                            'stuff': {
                                'too': True,
                            },
                        },
                    },
                },
            },
            'put_1': 'test 1',
            'put_2': 'test {a}',
        }, self._eval("""
            # set variables of with values of every type
            $null = null
            $a = 1
            $b = true
            $c = "Test"
            $c1 = "Test {c}"
            $c2 = 'Test {c}'
            $d = 3.14159
            $e = [1, true, "Test", 3.14159, [1, true, "Test", 3.14159], {ok:true}]
            $f = {
                ok: true,
            }
            $g, $h, $i = "g", "h", "i"
            $j, $k = "j"
            $l, $m = ["l", "m", "n"]
            $o, $p = ["o"]
            $q, _, $s = ["q", "r", "s"]
            $t = $f.ok
            $u = $f.nonexistent
            $v, $w, $x, $y, $z = [1, true, "Test", 3.14159, [1, true, "Test", 3.14159]]

            $z0 = $z[0]
            $z1 = $z[1]
            $z2 = $z[2]
            $z3 = $z[3]

            # capture command results as variables, and also put a bunch of them on the same line
            put 1 -> $aa; put true -> $bb; put "Test" -> $cc; put 3.14159 -> $dd
            put {
                ok: true,
                always: {
                    finishing: {
                        each_others: "sentences",
                    },
                },
            } -> $ee

            $ee1, $ee2 = $ee.always, $ee.always.finishing
            $ee3, $ee4 = [$ee.always.finishing.each_others, $ee.always.finishing.each_others.sandwiches]

            $ee5 = $ee['ok']
            $ee6 = $ee['always'].finishing['each_others']
            $ekey = 'always'
            $ee7 = $ee[$ekey].finishing['each_others']
            $ee8 = $ee
            $ee8.always['finishing'].each_others = 'sandwiches'
            $ee8.always['finishing'].other['stuff'].too = True

            put "test {a}" -> $put_1
            put 'test {a}' -> $put_2
        """))

    def test_if_scopes(self):
        self.assertEqual({
            'a':             'top_a',
            'b':             'top_b',
            'a_if':          'top_a',
            'b_if':          'if_b',
            'a_if_if':       'if_if_a',
            'b_if_if':       'if_b',
            'a_after_if_if': 'top_a',
            'b_after_if_if': 'if_b',
            'a_after_if':    'top_a',
            'b_after_if':    'top_b',
            'enter_if_val':  51,
            'enter_el_val':  61,
            'result':        None,
        }, self._eval("""
            $a             = "top_a"
            $b             = "top_b"
            $a_if          = null
            $b_if          = null
            $a_if_if       = null
            $b_if_if       = null
            $a_after_if_if = null
            $b_after_if_if = null
            $a_after_if    = null
            $b_after_if    = null

            if $b = "if_b"; $b {
                $a_if = $a
                $b_if = $b

                if $a = "if_if_a"; $a {
                    $a_if_if = $a
                    $b_if_if = $b
                }

                $a_after_if_if = $a
                $b_after_if_if = $b
            }

            $a_after_if = $a
            $b_after_if = $b
            $enter_if_val = null
            $enter_el_val = null

            # if condition trigger, verify condition value, populate via assignment
            if $value = 51; $value > 50 {
                $enter_if_val = 51
            } else {
                $enter_if_val = 9999
            }

            # else condition trigger, verify condition value, populate via command output
            if put 61 -> $value; $value > 100 {
                $enter_el_val = 7777
            } else {
                $enter_el_val = 61
            }

            $result = null
        """))

    def test_syntax_errors(self):
        with self.assertRaises(parser.exceptions.ScriptError):
            self._eval("nonexistent")

    def test_conditionals(self):
        self.assertEqual({
            'ten':            10,
            'unset':          None,
            'true':           True,
            'false':          False,
            'string':         "string",
            'names':          ["Bob", "Steve", "Fred"],
            'if_eq':          True,
            'if_ne':          True,
            'if_eq_null':     True,
            'if_true':        True,
            'if_gt':          True,
            'if_gte':         True,
            'if_lt':          True,
            'if_lte':         True,
            'if_in':          True,
            'if_not_in':      True,
            'if_match_1':     True,
            'if_match_2':     True,
            'if_match_3':     True,
            'if_not_match_1': True,
            'if_not_match_2': True,
            'if_not_match_3': True,
            'if_match_4':     True,
            'if_match_5':     True,
            'if_match_6':     True,
            'if_not_match_4': True,
            'if_not_match_5': True,
            'if_not_match_6': True,
        }, self._eval("""
            $ten = 10
            $unset = null
            $true = true
            $false = false
            $string = "string"
            $names = ["Bob", "Steve", "Fred"]
            $if_eq = null
            $if_ne = null
            $if_eq_null = null
            $if_true = null
            $if_gt = null
            $if_gte = null
            $if_lt = null
            $if_lte = null
            $if_in = null
            $if_not_in = null
            $if_match_1 = null
            $if_match_2 = null
            $if_match_3 = null
            $if_match_4 = null
            $if_match_5 = null
            $if_match_6 = null
            $if_not_match_1 = null
            $if_not_match_2 = null
            $if_not_match_3 = null
            $if_not_match_4 = null
            $if_not_match_5 = null
            $if_not_match_6 = null

            if $ten == 10                    { $if_eq          = true }
            if $unset == null                { $if_eq_null     = true }
            if $ten != 5                     { $if_ne          = true }
            if $ten > 5                      { $if_gt          = true }
            if $ten >= 10                    { $if_gte         = true }
            if $ten < 20                     { $if_lt          = true }
            if $ten <= 10                    { $if_lte         = true }
            if $true                         { $if_true        = true }
            if not $false                    { $if_false       = true }
            if "Steve" in $names             { $if_in          = true }
            if "Bill" not in $names          { $if_not_in      = true }
            if $string =~ /str[aeiou]ng/     { $if_match_1     = true }
            if $string =~ /String/i          { $if_match_2     = true }
            if $string =~ /.*/               { $if_match_3     = true }
            if $string !~ /strong/i          { $if_not_match_1 = true }
            if $string !~ /String/           { $if_not_match_2 = true }
            if $string !~ /^ring$/           { $if_not_match_3 = true }
            if not $string !~ /str[aeiou]ng/ { $if_match_4     = true }
            if not $string !~ /String/i      { $if_match_5     = true }
            if not $string !~ /.*/           { $if_match_6     = true }
            if not $string =~ /strong/i      { $if_not_match_4 = true }
            if not $string =~ /String/       { $if_not_match_5 = true }
            if not $string =~ /^ring$/       { $if_not_match_6 = true }
        """))

    def test_expressions(self):
        self.assertEqual({
            'a': 2,
            'b': 6,
            'c': 20,
            'd': 5,
            'aa': 2,
            'bb': 6,
            'cc': 20,
            'dd': 5,
            # 'f': u'This 2 is {b} and done',
            'put_a': '    this is some stuff',
            'put_b': '    buncha\n    muncha\n    cruncha\n    lines',
        }, self._eval("""
            $a = 1 + 1
            $b = 9 - 3
            $c = 5 * 4
            $d = 50 / 10
            # $e = 4 * -6 * (3 * 7 + 5) + 2 * 7

            $aa = 1
            $aa += 1

            $bb = 9
            $bb -= 3

            $cc = 5
            $cc *= 4

            $dd = 50
            $dd /= 10

            # $f = "This {a}" + ' is {b}' + " and done"

            put begin
                this is some stuff
            end -> $put_a

            put begin
                buncha
                muncha
                cruncha
                lines
            end -> $put_b
        """))

    def test_loops_all_sorts_of_loops(self):
        self.assertEqual({
            'forevers':        9,
            'double_break':    [4, 1],
            'double_continue': [8, 9],
            'iterations':      4
        }, self._eval("""
            $forevers = 0
            $double_break = null
            $double_continue = null
            $iterations = null
            $things = [1,2,3,4,5]

            loop {
                if not $index < 10 {
                    break
                }

                $forevers = $index
            }

            loop $x in $things {
                $iterations = $index
            }

            loop count 10 {
                $topindex = $index

                loop count 10 {
                    if $topindex == 4 {
                        if $index == 2 {
                            break 2
                        }
                    }

                    $double_break = [$topindex, $index]
                }
            }

            loop count 10 {
                $topindex = $index

                loop count 10 {
                    if $topindex == 9 {
                        if $index >= 0 {
                            continue 2
                        }
                    }

                    $double_continue = [$topindex, $index]
                }
            }

            unset $things
        """))



"""
Implementation of the Chrome Remote DevTools debugging protocol.

See: https://chromedevtools.github.io/devtools-protocol
"""
from __future__ import absolute_import
from webfriend.rpc.event import Event
from uuid import uuid4
from collections import OrderedDict
import logging


class Base(object):
    supports_events = True
    domain = None

    def __init__(self, tab):
        if self.domain is None:
            raise ValueError("Cannot instantiate an RPC proxy without a domain class property.")

        self.tab       = tab
        self.callbacks = {}

    def initialize(self):
        pass

    def call(self, method, expect_reply=True, reply_timeout=None, **params):
        return self.tab.rpc(
            '{}.{}'.format(self.domain, method),
            expect_reply=expect_reply,
            reply_timeout=reply_timeout,
            **params
        )

    def enable(self):
        if self.supports_events:
            self.call('enable')

    def disable(self):
        if self.supports_events:
            self.call('disable')

    def call_boolean_response(self, method, field='result', **kwargs):
        if self.call(method, **kwargs).get(field) is True:
            return True

        return False

    def on(self, method, callback):
        # normalize method name
        if not method.startswith(self.domain + '.'):
            method = '{}.{}'.format(self.domain, method)

        # create handler dict if we need to
        if method not in self.callbacks:
            self.callbacks[method] = OrderedDict()

        callback_id = '{}.event_{}'.format(self.domain, uuid4())
        self.callbacks[method][callback_id] = callback

        logging.debug('Registered event handler {} for event {}'.format(
            callback_id,
            method
        ))

        return callback_id

    def remove_handler(self, callback_id):
        for _, callbacks in self.callbacks.items():
            for id, _ in callbacks.items():
                if callback_id == id:
                    del callbacks[callback_id]
                    return True

        return False

    def trigger(self, method, payload=None):
        event = Event(self, method, payload)

        if str(event) in self.callbacks:
            for callback_id, callback in self.callbacks[str(event)].items():
                if callable(callback):
                    response = callback(event)

                    if response is False:
                        break

        return event



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.parser import MetaModel, to_value, exceptions, types


class ConditionalExpression(MetaModel):
    def evaluate(self, environment, scope=None):
        if scope is None:
            scope = environment.scope

        # Inline Assignment and test condition
        # ------------------------------------------------------------------------------------------
        if self.assignment:
            self.assignment.assign(environment, scope, force=True)

            # recursively evaluate the condition portion of the statement
            return self.condition.evaluate(environment, scope)

        # Command Evaluation (with optional test condition)
        # ------------------------------------------------------------------------------------------
        elif self.command:
            # execute the command
            resultkey, result = environment.execute(self.command, scope)

            # whatever the result of the expression command was, put it in the calling scope
            scope.set(resultkey, result, force=True)

            # if the statement has a condition, then evaluate using it
            if self.condition:
                # recursively evaluate the condition portion of the statement
                return self.condition.evaluate(environment, scope)
            else:
                # no condition, so just do the same thing as a unary comparison
                return self.compare(result)

        # Comparison
        # ------------------------------------------------------------------------------------------
        else:
            result = self.compare(
                to_value(self.lhs, scope),
                self.operator,
                to_value(self.rhs, scope)
            )

        if self.negate is True:
            return not result
        else:
            return result

    def compare(self, lhs, operator=None, rhs=None):
        if operator is not None:
            if operator == '==':
                if rhs is None:
                    return (lhs is None)
                else:
                    return (lhs == rhs)

            elif operator == '!=':
                if rhs is None:
                    return (lhs is not None)
                else:
                    return (lhs != rhs)

            elif operator == '<':
                return (lhs < rhs)

            elif operator == '<=':
                return (lhs <= rhs)

            elif operator == '>':
                return (lhs > rhs)

            elif operator == '>=':
                return (lhs >= rhs)

            elif operator == '=~':
                if isinstance(rhs, types.RegularExpression):
                    return rhs.is_match('{}'.format(lhs))

                elif hasattr(rhs, 'match'):
                    return (rhs.match('{}'.format(lhs)) is not None)

                else:
                    raise ValueError(
                        "Match operator must specify a regular expression on the right-hand side"
                    )

            elif operator == '!~':
                if isinstance(rhs, types.RegularExpression):
                    return not rhs.is_match('{}'.format(lhs))

                elif hasattr(rhs, 'match'):
                    return (rhs.match('{}'.format(lhs)) is None)

                else:
                    raise ValueError(
                        "Unmatch operator must specify a regular expression on the right-hand side"
                    )

            elif operator == 'in':
                return (lhs in rhs)

            elif operator == 'notin':
                return (lhs not in rhs)

            else:
                raise exceptions.ScriptError("Unsupported operator '{}'".format(operator))

        elif lhs:
            return True

        return False


class IfElseBlock(MetaModel):
    def get_blocks(self, environment, scope=None):
        if self.if_expr.expression.evaluate(environment, scope=scope):
            return self.if_expr.blocks

        for elseif in self.elseif_expr:
            if elseif.statement.expression.evaluate(environment, scope=scope):
                return elseif.statement.blocks

        if self.else_expr:
            return self.else_expr.blocks

        return []



from __future__ import absolute_import
from webfriend.rpc import Base


class Inspector(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Inspector
    """
    domain = 'Inspector'



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.parser import to_value


class MetaModel(object):
    def __init__(self, parent, **kwargs):
        self.parent = parent

        for k, v in kwargs.items():
            if isinstance(v, list):
                v = tuple(v)

            setattr(self, k, v)


class Directive(MetaModel):
    pass


class LinearExecutionBlock(MetaModel):
    pass


class EventHandlerBlock(MetaModel):
    pass


class Expression(MetaModel):
    def process(self, scope, preserve_types=False):
        lhs = to_value(self.lhs, scope)
        op = self.operator

        if op is not None:
            rhs = to_value(self.rhs, scope)

            if op.power:
                return lhs ** rhs, None
            elif op.bitwise_not:
                return ~rhs, None
            elif op.multiply:
                return lhs * rhs, None
            elif op.divide:
                return lhs / rhs, None
            elif op.modulus:
                return lhs & rhs, None
            elif op.add:
                return lhs + rhs, None
            elif op.subtract:
                return lhs - rhs, None
            elif op.bitwise_and:
                return lhs & rhs, None
            elif op.bitwise_xor:
                return lhs ^ rhs, None
            elif op.bitwise_or:
                return lhs | rhs, None

            return None, None

        if preserve_types:
            return self.lhs, None
        else:
            return lhs, None



from __future__ import absolute_import
from __future__ import unicode_literals
from textx.metamodel import metamodel_from_str
import textx.exceptions
import textx
from webfriend.scripting.parser.grammar import FRIENDSCRIPT_GRAMMAR


def to_value(value, scope, preserve_strings=False):
    # expand variables into values first
    if isinstance(value, variables.Variable):
        value = value.resolve(scope)

    # evaluate expressions
    if isinstance(value, lang.Expression):
        value, _ = value.process(scope)

    # if we're told to, return string classes directly (for interpolation)
    if preserve_strings and isinstance(value, (types.StringLiteral, types.Heredoc)):
        return value

    # do type detection and extraction
    elif isinstance(value, types.Array):
        value = [to_value(v, scope) for v in value.values]

    elif isinstance(value, types.Object):
        value = dict([
            (k, to_value(v, scope)) for k, v in value.as_dict().items()
        ])

    elif isinstance(value, float):
        if int(value) == value:
            value = int(value)

    if isinstance(value, str):
        value = value.decode('UTF-8')

    # null become None
    if value == 'null':
        return None

    return value


from .lang import MetaModel  # noqa
from . import commands, conditions, exceptions, lang, loops, types, variables  # noqa


class Friendscript(object):
    def __init__(self, filename=None, data=None, environment=None):
        self.model = None
        self.data = None
        self.environment = environment

        try:
            self.metamodel = metamodel_from_str(
                FRIENDSCRIPT_GRAMMAR,
                classes=[
                    commands.CommandSequence,
                    commands.CommandStanza,
                    conditions.ConditionalExpression,
                    conditions.IfElseBlock,
                    lang.Directive,
                    lang.EventHandlerBlock,
                    lang.Expression,
                    lang.LinearExecutionBlock,
                    loops.FlowControlWord,
                    loops.LoopBlock,
                    types.Array,
                    types.Object,
                    types.RegularExpression,
                    variables.Assignment,
                    variables.CommandID,
                    variables.Variable,
                ],
                match_filters={
                    'StringLiteral':      types.StringLiteral,
                    'StringInterpolated': types.StringInterpolated,
                    'Heredoc':            types.Heredoc,
                },
            )

            if isinstance(data, basestring):
                self.loads(data)

            elif isinstance(filename, basestring):
                self.load_file(filename)

        except textx.exceptions.TextXSyntaxError as e:
            raise exceptions.SyntaxError(
                str(e),
                filename=filename,
                data=data,
                line=e.line,
                col=e.col
            )

        # and tell the model who we are
        self.model.manager = self

    def load_file(self, filename):
        self.model = self.metamodel.model_from_file(filename)
        self.data = open(filename).read()

    def loads(self, data):
        self.model = self.metamodel.model_from_str(data)
        self.data = data

    def blocks_by_type(self, metatypes):
        if isinstance(self.model, basestring):
            return []

        return [
            b for b in self.model.blocks if isinstance(b, metatypes)
        ]

    def get_item_position(self, item):
        if not hasattr(item, '_tx_position'):
            raise ValueError("Must specify a textx model object")

        return self.metamodel.parser.pos_to_linecol(item._tx_position)

    @property
    def blocks(self):
        return self.blocks_by_type(
            tuple(lang.LinearExecutionBlock._tx_inh_by)
        )

    @property
    def handlers(self):
        return self.blocks_by_type(lang.EventHandlerBlock)



"""
The `webfriend.scripting` module implements a simplistic, straightforward scripting environment
for automating various tasks within the Chromium web browser.

## Overview

The commands implemented in this
module provide simplified, opinionated implementations of common tasks such as loading URLs,
parsing and extracting information from the DOM, injecting and executing Javascript into pages, and
capturing web page screenshots.

## Scripting

Scripts consist of plain text files containing a sequence of commands that are executed in the
order they are written.  Return values of each command are stored after the command executes,
allowing subsequent commands to access useful details about the evaluation of previous commands.

If a command experiences an error the error is either raised as a fatal exception, causing the
script to terminate without evaluating the remaining commands, or the error is specified as
non-fatal.  Non-fatal errors are stored and are accessible to the remaining commands and can be
used to make decisions.
"""
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)  # noqa



from __future__ import absolute_import
import logging


class Reply(object):
    def __init__(self, reply, request=None, events=None):
        if not isinstance(reply, dict):
            raise AttributeError("Reply payload must be a dict")

        if 'id' not in reply:
            raise AttributeError("Reply payload must contain an 'id' field")

        if not isinstance(reply.get('result'), dict):
            reply['result'] = {}

        self.reply   = reply
        self.id      = reply['id']
        self.result  = reply['result']
        self.request = (request or {})
        self.events  = (events or [])

        logging.debug(' << [{:04d}] OK'.format(
            self.id
        ))

    def as_dict(self):
        return {
            'request':    self.request,
            'message_id': self.id,
            'reply':      self.reply,
            'events':     self.events,
        }

    def get(self, key, fallback=None):
        return self.result.get(key, fallback)

    def ievents(self, type=None):
        for event in self.events:
            if isinstance(type, basestring):
                if type != str(event):
                    continue

            yield event



from __future__ import absolute_import
from setuptools import setup, find_packages
from webfriend.info import version

setup(
    name='webfriend',
    description='An API binding and automation control language for the Chrome Remote Debugging API',
    version=version,
    author='Gary Hetzel',
    author_email='garyhetzel@gmail.com',
    url='https://github.com/ghetzel/webfriend',
    install_requires=[
        'click',
        'click-log',
        'colorlog',
        'ephemeral_port_reserve',
        'prompt_toolkit',
        'pygments',
        'requests',
        'termcolor',
        'textx',
        'urlnorm',
        'websocket-client',
    ],
    packages=find_packages(exclude=['*.tests']),
    entry_points={
        'console_scripts': [
            'webfriend = webfriend.cli:main',
        ],
    },
    classifiers=[],
)



from __future__ import absolute_import
from __future__ import unicode_literals


FRIENDSCRIPT_GRAMMAR = r'''
Friendscript:
    Shebang?
    blocks *= Block
;

Shebang:
    /^#!.*$/
;

Comment:
    ( /#.*$/ )
;

ReservedWord:
    (
        'as' |
        'break' |
        'case' |
        'continue' |
        'else' |
        'if' |
        'in' |
        'is' |
        'loop' |
        'not' |
        'on' |
        'unset' |
        'when'
    )
;

FlowControlWord:
    ( is_break?='break' levels=INT? | is_continue?='continue' levels=INT? )
;

Block:
    (FlowControlWord | EventHandlerBlock | LinearExecutionBlock ) ';'?
;

LinearExecutionBlock:
    ( Assignment | Directive | Expression | IfElseBlock | LoopBlock | CommandSequence )
;

Assignment:
    destinations+=Variable[','] op=AssignmentOperator sources+=Expression[',']
;

Directive:
    (
        is_unset?='unset' variables+=Variable[',']
    )
;

Operator:
    (
        power='**'      |
        multiply='*'    |
        divide='/'      |
        modulus='%'     |
        add='+'         |
        subtract='-'    |
        bitwise_and='&' |
        bitwise_or='|'  |
        bitwise_not='~' |
        bitwise_xor='^'
    )
;

AssignmentOperator:
    (
        eq='='          |
        multi_eq='*='   |
        div_eq='/='     |
        plus_eq='+='    |
        minus_eq='-='   |
        and_eq='&='     |
        or_eq='|='      |
        append='<<'
    )
;

Expression:
    lhs=ValueYielding (operator=Operator rhs=Expression)?
;

ValueYielding:
    ( Type | Variable )
;

ConditionalExpression:
    negate?='not'? (
        assignment=Assignment ';' condition=ConditionalExpression |
        command=Command ( ';' condition=ConditionalExpression )? |
        lhs=Expression operator=MatchOperator rhs=RegularExpression |
        lhs=Expression ( operator=ComparisonOperator rhs=Expression )?
    )
;

ComparisonOperator:
    ( '==' | '!=' | '>=' | '<=' | '<' | '>' | 'in' | 'not' 'in' )
;

MatchOperator:
    ( '!~' | '=~' )
;

IfElseBlock:
    if_expr      = IfStanza
    elseif_expr *= ElseIfStanza
    else_expr    = ElseStanza?
;

IfStanza:
    'if' expression=ConditionalExpression '{'
        blocks *= Block
    '}'
;

ElseIfStanza:
    'else' statement=IfStanza
;

ElseStanza:
    'else' '{'
        blocks *= Block
    '}'
;

LoopBlock:
    'loop' (
        fixedlen=LoopConditionFixedLength |
        iterable=LoopConditionIterable |
        bounded=LoopConditionBounded |
        truthy=LoopConditionTruthy
    )?
    '{'
        blocks *= Block
    '}'
;

LoopConditionIterable:
    variables+=Variable[','] 'in' ( iterator=Command | iterator=Variable )
;

LoopConditionBounded:
    initial=Command ';' termination=ConditionalExpression ';' next=Command
;

LoopConditionTruthy:
    termination=ConditionalExpression
;

LoopConditionFixedLength:
    'count' count=IntOrVariable
;

IntOrVariable:
    (Variable | NUMBER)
;

Variable:
    (
        '$'- parts+=VariableName['.'] |
        skip?='_'
    )
;

VariableName[noskipws]:
    key=/[^\d\W][\w]*\b/ ( '[' index=VariableIndex ']' )?
;

VariableIndex:
    ( Variable | INT | String )
;

EventHandlerBlock:
    'on' pattern=String? isolated?='isolated'? '{'
        blocks *= LinearExecutionBlock
    '}'
;

CommandSequence:
    commands += Command[/;|$/]
;

Command:
    name=CommandName (
        id=CommandID options=CommandStanza? |
        options=CommandStanza
    )? (
        '->' resultkey=ResultKeyType
    )?
;

ResultKeyType:
    (Variable | 'null')
;

CommandID:
    (variable=Variable | value=ScalarType)
;

CommandStanza:
    '{'
        options *= KeyValuePair[','] ','?
    '}'
;

KeyValuePair:
    name=KeyType ':' values+=Expression
;

KeyType:
    ( !ReservedWord | ID | String )
;

Type:
    ( Object | Array | RegularExpression | ScalarType | Null )
;

Null:
    'null'
;

ScalarType:
    ( NUMBER | BOOL | String )
;

String[noskipws]:
    ( StringLiteral | StringInterpolated | Heredoc )
;

StringLiteral:
    /'[^\']*'/
;

StringInterpolated:
    /"[^\"]*"/
;

Heredoc:
    /begin(.|\n)*?(?!\n\s*end\s+)end/
;

RegularExpression:
    '/'- pattern=/[^\/]+/ '/' options*=/[ilmsu]/
;

Object:
    '{'
        items *= KeyValuePair[','] ','?
    '}'
;

Array:
    '['
        values *= Expression[','] ','?
    ']'
;

CommandName:
    !ReservedWord ID+['::']
;
'''



from __future__ import absolute_import
from __future__ import unicode_literals


class Scope(object):
    """
    A Scope is an object that allows for a split view of keys that are readable from all associated
    parent scopes, but only settable locally.

    This means that when retrieving values, a
    non-existent key will propagate the get up to the scope's parent, and so on until a value is
    found or a scope without a parent throws a KeyError.

    Setting values, on the other hand, are always set locally and only locally.  Taken together,
    these two behaviors form a hierarchical data structure in which children can override the values
    of their parents without modifying the parent.

    This is used to track the state of variables at various levels during the execution of commands,
    event handlers, and routine definitions during script evaluation.

    Args:
        data (dict, optional):
            The initial data, if any, to populate the Scope with.

        parent (Scope, optional):
            The parent scope, if any, to inherit values from.
    """
    SCALAR_TYPES = (basestring, int, float, bool)
    nested_joiner = '.'

    def __init__(self, data=None, parent=None):
        if parent and not isinstance(parent, Scope):
            raise ValueError("Parent scope must also be a Scope")

        self.parent = parent
        self.data   = (data or {})
        self.modified_keys = set()

    @property
    def level(self):
        lvl = 1

        if self.parent:
            lvl += self.parent.level

        return lvl

    @classmethod
    def split_key(cls, key):
        if isinstance(key, tuple):
            if not len(key):
                raise KeyError("Must specify an index value")

            return key[0], key
        else:
            parts = key.split(cls.nested_joiner)

            for i, part in enumerate(parts):
                if isinstance(part, str):
                    parts[i] = part.decode('UTF-8')

            return parts[0], parts

    @classmethod
    def join_key(cls, key):
        if isinstance(key, (list, tuple)):
            return cls.nested_joiner.join(['{}'.format(k) for k in key])
        return '{}'.format(key)

    def ancestors(self):
        out = []

        if self.parent:
            out += self.parent.ancestors()

        out.append(self)

        return out

    def owner_of(self, key):
        top_key, _ = self.split_key(key)

        # if the key exists, but is not ours...
        if top_key in self and not self.is_local(top_key):
            # and if we have a parent...
            if self.parent:
                # then let the parent deal with it
                return self.parent.owner_of(key)

        return self

    def as_dict(self):
        out = {}

        if self.parent:
            out.update(self.parent.as_dict())

        out.update(self.data)
        return out

    def items(self):
        return self.as_dict().items()

    def is_local(self, key):
        top_key, _ = self.split_key(key)

        if top_key in self.data:
            return True

        return False

    def get(self, key, fallback=None):
        try:
            return self[key]
        except KeyError:
            return fallback

    def set(self, key, value, force=False, unset=False):
        # only unset if the key exists
        if unset and key not in self:
            return

        top_key, parts = self.split_key(key)

        if isinstance(value, str):
            value = value.decode('UTF-8')

        # force means that we always make ourself the owner
        if force:
            owner = self
        else:
            owner = self.owner_of(key)

        base = owner.data

        for k in parts[:-1]:
            if k not in base:
                base[k] = {}
            elif not isinstance(base[k], dict):
                raise ValueError(
                    "Cannot set intermediate key '{}': key exists, but not a dict (is: {})".format(
                        k,
                        base[k].__class__.__name
                    )
                )

            base = base[k]

        if unset:
            if parts[-1] in base:
                del base[parts[-1]]
        else:
            base[parts[-1]] = value

        if owner is self:
            self.modified_keys.add(self.join_key(key))

    def unset(self, key, force=False):
        self.set(key, None, force=force, unset=True)

    def update(self, other, force=False):
        if isinstance(other, Scope):
            other = other.as_dict()

        if not isinstance(other, dict):
            raise ValueError("update() value must be a Scope or dict")

        for k, v in other.items():
            self.set(k, v, force=force)

        return None

    def __len__(self):
        return len(self.as_dict())

    def __getitem__(self, key):
        top_key, parts = self.split_key(key)

        try:
            base = self.data

            # try to traverse our local data and retrieve the value
            # if that fails at any point, make our parent do the same, on and on
            # until there are no parents left and we just raise the KeyError
            for k in parts:
                if isinstance(base, (list, tuple)):
                    try:
                        base = base[int(k)]
                        continue
                    except ValueError as e:
                        if str(e).startswith('invalid literal for int'):
                            raise ValueError(
                                "Attempted to use non-numeric value '{}' to access an array element".format(k)
                            )
                        else:
                            raise

                # try to access key via the [] operator
                if hasattr(base, '__getitem__') and not isinstance(base, self.SCALAR_TYPES):
                    try:
                        base = base[k]
                        continue
                    except KeyError:
                        pass

                if isinstance(base, object) and hasattr(base, str(k)):
                    base = getattr(base, str(k))
                    continue

                raise KeyError(
                    "Cannot access key '{}' on {} value".format(k, base.__class__.__name__)
                )

            return base
        except KeyError:
            if self.parent:
                return self.parent[key]
            raise

    def __setitem__(self, key, value):
        self.set(key, value)

    def __delitem__(self, key):
        self.unset(key)

    def __contains__(self, key):
        _, parts = self.split_key(key)

        # traverse our local data and make sure we have the given key
        try:
            base = self.data

            for k in parts[:-1]:
                base = base[k]

            if parts[-1] in base:
                return True

        except KeyError:
            pass

        # if we got to this point, try our parent
        if self.parent:
            if key in self.parent:
                return True

        return False

    def __str__(self):
        return 'Scope<level={},variables={}>'.format(
            self.level,
            len(self.data)
        )



"""
The `webfriend` module provides access to the browser automation and debugging facilities of
Google Chrome using the built-in [Chrome DevTools](https://developer.chrome.com/devtools) interface.
"""
# flake8: noqa
from __future__ import absolute_import
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)  # noqa



# flake8: noqa
from __future__ import absolute_import
from pygments.lexer import RegexLexer, bygroups
from pygments.token import *


class FriendscriptLexer(RegexLexer):
    name = 'Friendscript'
    aliases = ['friendscript']
    filenames = ['*.fs']

    tokens = {
        'root': [
            (r'^#!.*\n', Comment.Hashbang),
            (r'#.*\n', Comment.Single),
            (r'\/.*\/[a-z]*', String.Regex),
            (r'\s+(=~|%|~|!~|<<|->|!=|(?:&|\||\^|\+|-|\*|\/|=|<|>)=?)\s+', bygroups(Operator)),
            (r'\s+(in|is|not)\s+', bygroups(Operator.Word)),
            (r'\s+(as|break|case|continue|else|if|loop(\s+count)?|on|unset|when)\s+', bygroups(Keyword)),
            (r'\d', Number),
            (r'(true|false|null)', Keyword.Pseudo),
            (r'\'', String.Single, 'str'),
            (r'"', String.Double, 'str-double'),
            (r'\$[^\d\W][\w\.]*\b', Name.Variable),
            (r'[\{\},\.\[\];]', Punctuation),
            (r'([\w\d]+)\s*(:)(?!:)', bygroups(Name.Attribute, Punctuation)),
            (r'([\w\d][\w\d]*)(::[\w\d][\w\d]*)*', Name.Function),
            (r'\s+', Text),
        ],
        'str': [
            (r'[^\']', String.Single),
            (r'\'', String.Single, '#pop'),
        ],
        'str-double': [
            (r'{[^}]*}', String.Interpol),
            (r'[^"]', String.Double),
            (r'"', String.Double, '#pop'),
        ],
    }



from __future__ import absolute_import
from webfriend.rpc import Base
from datetime import datetime
from base64 import b64decode


class Cookie(object):
    def __init__(self, rpc, definition):
        self._rpc          = rpc
        self._definition   = definition
        self.name          = definition['name']
        self.value         = definition['value']
        self.domain        = definition.get('domain')
        self.path          = definition.get('path')
        self.expires_epoch = definition.get('expires')
        self.size          = definition.get('size')
        self.http_only     = definition.get('httpOnly')
        self.secure        = definition.get('secure')
        self.session       = definition.get('session')
        self.same_site     = definition.get('sameSite')

        if self.expires_epoch is not None:
            self.expires   = datetime.fromtimestamp(self.expires_epoch / 1e3)

    def as_dict(self):
        return dict([
            (k, v) for k, v in self.__dict__.items() if not k.startswith('_')
        ])


class Network(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Network
    """
    domain = 'Network'

    connection_types = [
        'none', 'cellular2g', 'cellular3g', 'cellular4g', 'bluetooth', 'ethernet', 'wifi', 'wimax',
        'other'
    ]

    def set_user_agent(self, user_agent):
        self.call('setUserAgentOverride', userAgent=user_agent)

    def set_headers(self, headers):
        if not isinstance(headers, dict):
            raise AttributeError("Headers must be specified as a dict")

        self.call('setExtraHTTPHeaders', headers=headers)

    def get_response_body(self, request_id):
        reply = self.call('getResponseBody', requestId=request_id)
        body = reply.get('body')

        if not body:
            return None

        if reply.get('base64Encoded') is True:
            body = b64decode(body)

        return body

    @property
    def can_clear_browser_cache(self):
        return self.call_boolean_response('canClearBrowserCache')

    @property
    def can_clear_browser_cookies(self):
        return self.call_boolean_response('canClearBrowserCookies')

    @property
    def can_emulate_network_conditions(self):
        return self.call_boolean_response('canEmulateNetworkConditions')

    def clear_browser_cache(self):
        self.call('clearBrowserCache')

    def clear_browser_cookies(self):
        self.call('clearBrowserCookies')

    def emulate_network_conditions(
        self,
        offline=False,
        latency_ms=0,
        throughput_down_bps=None,
        throughput_up_bps=None,
        connection_type=None,
    ):
        params = {
            'offline': offline,
            'latency':            latency_ms,
            'downloadThroughput': 10e9,
            'uploadThroughput':   10e9,
        }

        if connection_type is not None:
            if connection_type not in self.connection_types:
                raise AttributeError("Connection Type must be one of: {}".format(
                    ', '.join(self.connection_types)
                ))

            params['connectionType'] = connection_type

        self.call('emulateNetworkConditions', **params)

    def disable_cache(self):
        self.call('setCacheDisabled', cacheDisabled=True)

    def enable_cache(self):
        self.call('setCacheDisabled', cacheDisabled=False)

    def set_blocked_urls(self, urls):
        if not isinstance(urls, list):
            raise AttributeError("Blocked URLs must be a list")
        self.call('setBlockedURLs', urls=urls)

    def replay_xhr(self, request_id):
        self.call('replayXHR', requestId=request_id)

    def get_cookies(self, urls=None):
        if isinstance(urls, list):
            reply = self.call('getCookies', urls=urls)
        else:
            reply = self.call('getAllCookies')

        return [
            Cookie(self, c) for c in reply.get('cookies', [])
        ]

    def delete_cookie(self, url, name):
        self.call('deleteCookie', cookieName=name, url=url)

    def set_cookie(
        self,
        url,
        name,
        value,
        domain=None,
        path=None,
        secure=None,
        http_only=None,
        same_site=None,
        expires=None
    ):
        """
        Create or update a cookie based on the given values.
        """
        params = {
            'url':   url,
            'name':  name,
            'value': value,
        }

        if domain is not None:
            params['domain'] = domain

        if path is not None:
            params['path'] = path

        if isinstance(secure, bool):
            params['secure'] = secure

        if isinstance(http_only, bool):
            params['httpOnly'] = http_only

        if isinstance(same_site, basestring):
            params['sameSite'] = same_site

        if isinstance(expires, int):
            params['expirationDate'] = expires
        elif isinstance(expires, datetime):
            params['expirationDate'] = int(expires.strftime('%s'))

        return self.call_boolean_response(self.call('setCookie', **params), 'success')

    def get_certificate(self, origin):
        # getCertificate
        return Exception("NOT IMPLEMENTED")



from __future__ import unicode_literals
from webfriend.scripting.scope import Scope
import inspect
import importlib


class CommandProxy(object):
    qualifier = None
    doc_name = None
    default_qualifier = 'core'
    default_method_name = 'default'

    enabled_proxies = [
        'cookies',
        'core',
        'file',
        'fmt',
        'page',
        'utils',
        'vars',
    ]

    def __init__(self, browser, environment=None, scope=None):
        self.browser = browser
        self.environment = environment
        self.scope   = (scope or Scope())
        self._tab    = None

    @property
    def tab(self):
        if self._tab is None:
            return self.browser.default
        else:
            try:
                return self.browser.tabs[self._tab]
            except KeyError:
                raise KeyError("Could not retrieve unknown tab '{}'".format(self._tab))

    def __contains__(self, local_command_name):
        try:
            if local_command_name == self.as_qualifier():
                return callable(getattr(self, self.default_method_name))
            else:
                return callable(getattr(self, local_command_name))
        except AttributeError:
            pass

        return False

    def __getitem__(self, local_command_name):
        if local_command_name == self.as_qualifier():
            return getattr(self, self.default_method_name)
        else:
            return getattr(self, local_command_name)

    def has_default(self):
        return (self.as_qualifier() in self)

    @classmethod
    def get_all_proxies(cls):
        proxies = []
        for proxy_module in cls.enabled_proxies:
            module = importlib.import_module('webfriend.scripting.commands.{}'.format(
                proxy_module
            ))

            members = inspect.getmembers(module, inspect.isclass)

            for subcls in members:
                if issubclass(subcls[1], CommandProxy) and subcls[1] is not CommandProxy:
                    proxies.append((subcls[1].as_qualifier(), subcls[1]))

        return proxies

    @classmethod
    def get_command_names(cls):
        names = set()

        for name in dir(cls):
            # underscored items are not commands
            if name.startswith('_'):
                continue

            # methods inherited from CommandProxy are not commands
            if name in dir(CommandProxy):
                continue

            if callable(getattr(cls, name)):
                if len(cls.as_qualifier()):
                    if cls.as_qualifier() == cls.default_qualifier:
                        names.add(name)
                    elif name == cls.default_method_name:
                        names.add(cls.as_qualifier())
                    else:
                        names.add(cls.qualify(name))

        return names

    @classmethod
    def qualify(cls, name):
        if name == cls.as_qualifier():
            return name
        else:
            return '{}::{}'.format(cls.as_qualifier(), name)

    @classmethod
    def as_qualifier(cls):
        if cls.qualifier is not None:
            return cls.qualifier

        return cls.__name__.split('.')[-1].replace('Proxy', '').lower()



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.parser import MetaModel, to_value
import copy


class Variable(MetaModel):
    def as_key(self, scope):
        parts = tuple()

        for part in self.parts:
            if len(part.key):
                parts = parts + (part.key.decode('UTF-8'),)

            if part.index is not None:
                index = to_value(part.index, scope)

                if index is not None:
                    parts = parts + (index,)

        return parts

    def resolve(self, scope, fallback=None):
        if self.skip:
            return fallback

        value = scope.get(self.as_key(scope), fallback)

        return to_value(value, scope)


class CommandID(MetaModel):
    def resolve(self, scope):
        if self.variable:
            return self.variable.resolve(scope)

        return to_value(self.value, scope)


class Assignment(MetaModel):
    def assign(self, environment, scope, force=False):
        sources_consumed = 0
        sources = self.sources

        # provide support for unpacking lists and tuples into multiple variables
        if len(self.destinations) > 1:
            if len(self.sources) == 1:
                first_source = environment.interpolate(self.sources[0], scope=scope)

                if isinstance(first_source, (list, tuple)):
                    sources = first_source

        # for each source value
        for i, source in enumerate(sources):
            # if there is a destination for it
            if i < len(self.destinations):
                destination = self.destinations[i]

                if not destination.skip:
                    source = environment.interpolate(source, scope=scope)

                    # assignments are BY VALUE (e.g.: always incur a copy)
                    if self.op.eq:
                        value = copy.copy(source)

                    elif self.op.append:
                        value = scope.get(destination.as_key(scope))

                        if not isinstance(value, list):
                            value = [value]

                        value.append(copy.copy(source))
                    else:
                        value = scope.get(destination.as_key(scope), 0)
                        rhs = copy.copy(source)

                        if self.op.multi_eq:
                            value = value * rhs
                        elif self.op.div_eq:
                            value = value / rhs
                        elif self.op.plus_eq:
                            value = value + rhs
                        elif self.op.minus_eq:
                            value = value - rhs
                        elif self.op.and_eq:
                            value = value * rhs
                        elif self.op.or_eq:
                            value = value | rhs

                    scope.set(
                        destination.as_key(scope),
                        value,
                        force=force
                    )

                sources_consumed += 1

        # null out the remaining destinations
        if len(self.destinations) > sources_consumed:
            for unset_destination in self.destinations[sources_consumed:]:
                if not unset_destination.skip:
                    scope.set(
                        unset_destination.as_key(scope),
                        None,
                        force=force
                    )



from __future__ import absolute_import
from __future__ import unicode_literals
import json
import logging
import re
from webfriend.scripting.environment import Environment
from webfriend.scripting.scope import Scope
from webfriend.scripting.execute import execute_script
from prompt_toolkit import prompt
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.layout.lexers import PygmentsLexer
from prompt_toolkit.styles import style_from_pygments
from prompt_toolkit.token import Token
from webfriend.scripting.parser.pygments import FriendscriptLexer
from pygments.styles.monokai import MonokaiStyle


class FriendscriptCompleter(Completer):
    def __init__(self, commands, *args, **kwargs):
        self.commands = commands
        super(FriendscriptCompleter, self).__init__(*args, **kwargs)

    def get_completions(self, document, complete_event):
        line = document.current_line.lstrip()

        for command in self.commands:
            if command.startswith(line):
                yield Completion(command, start_position=len(line) - len(command))


class REPL(object):
    INTERNAL_COMMANDS = [
        'help',
    ]

    prompt = '(friendscript) '

    def __init__(self, browser, environment=None):
        self.browser = browser

        if environment:
            self.environment = environment
        else:
            self.environment = Environment(Scope(), browser=self.browser)

        self.completer = FriendscriptCompleter(
            sorted(self.environment.get_command_names())
        )
        self.history = InMemoryHistory()
        self.last_result = None
        self.echo_results = True

    def get_last_result(self, cli):
        toolbar = []

        if isinstance(self.last_result, Exception):
            toolbar.append((Token.Error, str(self.last_result)))

        return toolbar

    def run(self):
        while True:
            line = prompt(
                '{}'.format(self.prompt),
                lexer=PygmentsLexer(FriendscriptLexer),
                style=style_from_pygments(MonokaiStyle, {
                    Token.Error: '#ansiwhite bg:#ansired',
                }),
                completer=self.completer,
                history=self.history,
                get_bottom_toolbar_tokens=self.get_last_result
            )

            try:
                self.last_result = execute_script(
                    self.browser,
                    line,
                    environment=self.environment,
                    preserve_state=True
                )

                self.environment.set_scope(self.last_result)

                if self.echo_results and not self.is_internal_command(line):
                    print(json.dumps(self.last_result.as_dict(), indent=4))

            except Exception as e:
                self.last_result = e
                logging.exception('FAIL')
                print(e)

    def is_internal_command(self, line):
        parts = re.split(r'\s+', line)

        if len(parts) and parts[0] in self.INTERNAL_COMMANDS:
            return True

        return False



from __future__ import absolute_import
from __future__ import unicode_literals
from collections import OrderedDict
from webfriend.scripting.parser import MetaModel
import re


class Array(MetaModel):
    pass


class String(unicode):
    def __new__(cls, value, *args, **kwargs):
        if not isinstance(value, unicode):
            value = value.encode('UTF-8')

        value = cls.preprocess(value)

        return unicode.__new__(cls, value, *args, **kwargs)

    @classmethod
    def preprocess(cls, value):
        return value


class StringLiteral(String):
    rx_start = re.compile('^\'')
    rx_end   = re.compile('\'$')

    @classmethod
    def preprocess(cls, value):
        value = cls.rx_start.sub('', value)
        value = cls.rx_end.sub('', value)
        return value


class StringInterpolated(String):
    rx_start = re.compile('^\"')
    rx_end   = re.compile('\"$')

    @classmethod
    def preprocess(cls, value):
        value = cls.rx_start.sub('', value)
        value = cls.rx_end.sub('', value)
        return value


class Heredoc(String):
    rx_start = re.compile('^begin(\s*\n)?')
    rx_end   = re.compile('\s+end$')

    @classmethod
    def preprocess(cls, value):
        value = cls.rx_start.sub('', value)
        value = cls.rx_end.sub('', value)
        return value

    def deindent(self, n):
        return '\n'.join([
            line.replace(' ' * n, '', 1) for line in self.split('\n')
        ])


class RegularExpression(MetaModel):
    def __init__(self, parent, **kwargs):
        super(RegularExpression, self).__init__(parent, **kwargs)

        if not len(self.pattern):
            raise ValueError("Must specify a pattern")

        self.rx = re.compile(self.pattern, flags=self.get_flags())

    def get_flags(self):
        flags = 0

        for option in self.options:
            if option == 'i':
                flags |= re.IGNORECASE
            elif option == 'l':
                flags |= re.LOCALE
            elif option == 'm':
                flags |= re.MULTILINE
            elif option == 's':
                flags |= re.DOTALL
            elif option == 'u':
                flags |= re.UNICODE

        return flags

    def is_match(self, value):
        if self.rx.match(value):
            return True

        return False

    def match(self, value):
        return self.rx.match(value)

    def sub(self, repl, value):
        return self.rx.sub(repl, value)


class Object(MetaModel):
    def __init__(self, parent, **kwargs):
        self.items = []
        self._data  = OrderedDict()

        super(Object, self).__init__(parent, **kwargs)

        for kv in self.items:
            if not len(kv.values):
                continue

            if len(kv.values) == 1:
                self._data[kv.name] = kv.values[0]
            else:
                self._data[kv.name] = kv.values

    def to_json(self):
        return self.as_dict()

    def as_dict(self):
        return self._data

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]
